/****************************************************/
/*  Final Project                                   */
/*  ocean.c                                         */
/*  Tzu-Kai Chou                                    */
/****************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>

#define NumInterval 100
#define NumFish     30

#define X_deep 10
#define Y_wide 10

#define nothing 0;
#define fish    1;

#define sea      0;
#define boundary 1;

#define N  1
#define NE 2
#define E  3
#define SE 4
#define S  5
#define SW 6
#define W  7
#define NW 8

void main(int argc,char *argv[])
{
    int ProcRank,ProcSize;
    
    int i,j,k,l,m;
    int randomX,randomY,randomRank;
 

    MPI_Comm Comm2D;
    int GridDim[2]={1,1};
    int WrapAround[2]={1,1};
    int Coords[2];    

    int LeftNeighbor,RightNeighbor,BottomNeighbor,TopNeighbor;
    int CountStart=0,CountEnd=0,TotalStart=0,TotalEnd=0,Count=0;

    MPI_Datatype contig,strided;

    MPI_Status status;

    FILE *fp;
    char str[81];

    int  MyRandom(int LowerBound,int UpperBound);

    void MySetGrid(int p,int dim[]);

    void Swimming(int x,int y,int xb,int yb,
                  int spc[X_deep+2][Y_wide+2],
                  int mk [X_deep+2][Y_wide+2] );

    void ClearRingGuard( int spc[X_deep+2][Y_wide+2] ); 

    void CollisionChk( int p_spc[X_deep+2][Y_wide+2],
                       int spc  [X_deep+2][Y_wide+2],
                       int mk   [X_deep+2][Y_wide+2]  );

    void Spawn(int spc[X_deep+2][Y_wide+2]);
    void   Die(int spc[X_deep+2][Y_wide+2]);
     
    int Species [X_deep+2][Y_wide+2];
    int Mask    [X_deep+2][Y_wide+2];
    int PrevSpc [X_deep+2][Y_wide+2];

/*******************************************************************/
/*  MPI Initialization                                             */
/*******************************************************************/
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &ProcSize);
    MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);

/*******************************************************************/
/*  Determine Grid Dimention sizes                                 */
/*  Create Cartesian Grid                                          */
/*******************************************************************/
    MySetGrid(ProcSize,GridDim);
    MPI_Cart_create(MPI_COMM_WORLD,2,GridDim,WrapAround,0,&Comm2D);
    MPI_Cart_coords(Comm2D,ProcRank,2,Coords);

    MPI_Cart_shift(Comm2D,0,1,&BottomNeighbor,&TopNeighbor);
    MPI_Cart_shift(Comm2D,1,1,&LeftNeighbor,&RightNeighbor);

/*******************************************************************/
/*  MPI define type (contig for row, strided for column)           */
/*  Grouped MSG for communication                                  */
/*******************************************************************/
    MPI_Type_contiguous(Y_wide,MPI_INT,&contig);
    MPI_Type_vector(X_deep,1,Y_wide+2,MPI_INT,&strided);
    MPI_Type_commit(&contig);
    MPI_Type_commit(&strided);

/*******************************************************************/
/*  2D Ocean Initialization                                        */
/*******************************************************************/
    for(i=0;i<X_deep+2;i++)
      for(j=0;j<Y_wide+2;j++)
      {
         Species [i][j]=nothing;
         Mask    [i][j]=sea;
      }

    /* Set up boundary */
    if(Coords[1]==0)
       for(i=1; i<Y_wide+1; i++)
           Mask[i][1]=boundary;

    if(Coords[1]==(GridDim[1]-1))
       for(i=1; i<Y_wide+1; i++)
           Mask[i][X_deep]=boundary;
       
    if(Coords[0]==0)  
       for(j=1; j<X_deep+1; j++)
           Mask[1][j]=boundary;

    if(Coords[0]==(GridDim[0]-1))
       for(j=1; j<X_deep+1; j++)
           Mask[Y_wide][j]=boundary;

/*******************************************************************/
/*  Aquariums Initialization                                       */
/*******************************************************************/
    for(i=0;i<NumFish;i++)
    {
       randomRank=MyRandom(0,ProcSize-1);
       if(ProcRank==randomRank)
       { 
          while(1){
             randomX=MyRandom(1,X_deep);
             randomY=MyRandom(1,Y_wide);
             if(Species[randomX][randomY]==0) /*0=nothing there*/
             {
                Species[randomX][randomY]=fish;
                break;
             } 
          }
       }
    }

    for(i=1;i<=X_deep;i++)
      for(j=1;j<=Y_wide;j++)
        if(Species[i][j]==1)
           CountStart+=1;
    MPI_Reduce(&CountStart,&TotalStart,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);

/*******************************************************************/
/*  Fish swimming,breeding in the 2D ocean                         */
/*******************************************************************/
    for(k=0;k<NumInterval;k++)
    {
       for(i=1;i<=X_deep;i++)
         for(j=1;j<=Y_wide;j++)
         {
            if(Species[i][j]==1) 
               Swimming(i,j,X_deep,Y_wide,Species,Mask);
         }
 
       /*save status(before send/receive)*/
       for(i=0;i<X_deep+2;i++)
         for(j=0;j<Y_wide+2;j++)
            PrevSpc[i][j]=Species[i][j];

       /* Communicate edge rows and columns to neighboring */
       /* processors to put in their guard rings.          */
       MPI_Sendrecv(&(Species[1][0]),1,strided,LeftNeighbor,30, 
		    &(Species[1][Y_wide]),1,strided,RightNeighbor,30,
		    Comm2D,&status);
       MPI_Sendrecv(&(Species[1][Y_wide+1]),1,strided,RightNeighbor,31,
		    &(Species[1][1]),1,strided,LeftNeighbor,31, 
		    Comm2D,&status);
       MPI_Sendrecv(&(Species[0][1]),1,contig,TopNeighbor,40, 
		    &(Species[X_deep][1]),1,contig,BottomNeighbor,40,
                    Comm2D,&status);
       MPI_Sendrecv(&(Species[X_deep+1][1]),1,contig,BottomNeighbor,41,
		    &(Species[1][1]),1,contig,TopNeighbor,41, 
		    Comm2D,&status);              

       CollisionChk(PrevSpc,Species,Mask);
       
       ClearRingGuard(Species);    
    
       Spawn(Species);
      
       Die(Species);   

       CountEnd=0; TotalEnd=0;
       for(i=1;i<=X_deep;i++)
         for(j=1;j<=Y_wide;j++)
           if(Species[i][j]==1)
              CountEnd+=1;
       MPI_Reduce(&CountEnd,&TotalEnd,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
   
       Count=Count+1;
       if((TotalEnd==0)||(TotalEnd==ProcSize*X_deep*Y_wide))
          break;
    }

    

/*******************************************************************/
/*  STD Output                                                     */
/*******************************************************************/
    if(ProcRank==0)    
       printf("StartFish=%d  EndFish=%d Intervals=%d\n",TotalStart,TotalEnd,Count);
    printf("Processor%d\n",ProcRank);
    for(i=1;i<=X_deep;i++)
    {
      for(j=1;j<=Y_wide;j++)      
         printf("%d ",Species[i][j]);
      printf("\n");
    }
/*******************************************************************/
/*  Finalization                                                   */
/*******************************************************************/
    MPI_Type_free(&contig);
    MPI_Type_free(&strided);
    MPI_Finalize();

}


/*******************************************************************/
/*  Random Number Generator                                        */
/*******************************************************************/
int MyRandom(int LowerBound,int UpperBound)
{
    int result;
    while(1){
       result=rand()%(UpperBound+1);
       if(result>=LowerBound)
          break;
    }
    return result;
}

/*******************************************************************/
/*  Description : This function determine the dimention sizes.     */
/*        Input : Number of process,Arry of dimention sizes        */
/*       Output : N/A                                              */
/*                                                                 */
/*      example :    p       SIZE                                  */
/*                   1       1 X 1                                 */
/*                   2       1 X 2                                 */
/*                   3       1 X 3                                 */
/*                   4       2 X 2                                 */
/*                   5       1 X 5                                 */
/*                   6       3 X 2                                 */
/*                   7       1 X 7                                 */
/*                   8       4 X 2                                 */
/*******************************************************************/
void MySetGrid(int p, int dim[])
{
    int i,temp;
    for(i=1;i<=p/2;i++)
    {
        if((p%i)==0)
           dim[0]=i;
    }
    dim[1]=p/dim[0];
    if(dim[0]>dim[1])
    {  temp=dim[0]; dim[0]=dim[1]; dim[1]=temp;}
}

/*******************************************************************/
/*  MoveTo                                                         */
/*  Ary[X-1][Y-1]  Ary[X-1][Y]  Ary[X-1][Y+1]                      */
/*  Ary[X]  [Y-1]  Ary[X]  [Y]  Ary[X][Y+1]                        */
/*  Ary[X+1][Y-1]  Ary[X+1][Y]  Ary[X+1][Y+1]                      */
/*******************************************************************/
void MoveTo(int dir,int x,int y,int spc[X_deep+2][Y_wide+2]) 
{
   switch(dir){
         case  N:if(spc[x-1][y]==0) /*0=nothing*/
                 {  spc[x-1][y]=spc[x][y]; spc[x][y]=nothing; }
                 break;
         case NE:if(spc[x-1][y+1]==0) /*0=nothing*/
                 {  spc[x-1][y+1]=spc[x][y]; spc[x][y]=nothing; }
                 break;
         case  E:if(spc[x][y+1]==0) /*0=nothing*/
                 {  spc[x][y+1]=spc[x][y]; spc[x][y]=nothing;   }
                 break;
         case SE:if(spc[x+1][y+1]==0) /*0=nothing*/
                 {  spc[x+1][y+1]=spc[x][y]; spc[x][y]=nothing; }
                 break;
         case  S:if(spc[x+1][y]==0) /*0=nothing*/
                 {  spc[x+1][y]=spc[x][y]; spc[x][y]=nothing; }
                 break;
         case SW:if(spc[x+1][y-1]==0) /*0=nothing*/
                 {  spc[x+1][y-1]=spc[x][y]; spc[x][y]=nothing; }
                 break;
         case  W:if(spc[x][y-1]==0) /*0=nothing*/
                 {  spc[x][y-1]=spc[x][y]; spc[x][y]=nothing; }   
                 break;
         case NW:if(spc[x-1][y-1]==0) /*0=nothing*/
                 {  spc[x-1][y-1]=spc[x][y]; spc[x][y]=nothing; }
                 break;
   } 
}

/*******************************************************************/
/*  Swimming                                                       */
/*  NW(8) N(1) NE(2)                                               */
/*  W (7)      E (3)                                               */
/*  SW(6) S(5) SE(4)                                               */
/*******************************************************************/
void Swimming(int x,int y,int xb,int yb,
              int spc[X_deep+2][Y_wide+2],
              int mk [X_deep+2][Y_wide+2] )
{
   int direction;
   if(mk[x][y]!=1) /*1=boundary(fish not on boundary)*/
   {
      direction=MyRandom(N,NW);   /*1-8*/
      MoveTo(direction,x,y,spc);  
   }
   else  /*fish on boundary*/
   {
      if((x==1)&&(y==1)) /*LeftTopCorner*/
      {
         direction=MyRandom(E,S);         /*       E(3) */
         MoveTo(direction,x,y,spc);       /* S(5) SE(4) */
      }
      if((x==1)&&(y==yb)) /*RightTopCorner*/
      {
         direction=MyRandom(S,W);         /* W (7)      */
         MoveTo(direction,x,y,spc);       /* SW(6) S(5) */
      } 
      if((x==xb)&&(y==1)) /*LeftBottomCorner*/
      {
         direction=MyRandom(N,E);         /* N(1) NE(2) */
         MoveTo(direction,x,y,spc);       /*      E (3) */
      } 
      if((x==xb)&&(y==yb)) /*RightBottomCorner*/
      {
         direction=MyRandom(1,3);         /* NW(8) N(1) */
         if(direction!=1)                 /* W (7)      */
            direction=direction+5;
         MoveTo(direction,x,y,spc);
      }
      if((x==1)&&(y!=1)&&(y!=yb)) /*TopBoundary*/
      {
         direction=MyRandom(E,W);         /* W (7)       E(3) */
         MoveTo(direction,x,y,spc);       /* SW(6) S(5) SE(4) */
      }
      if((x==xb)&&(y!=1)&&(y!=yb)) /*BottomBoundary*/
      {
         direction=MyRandom(1,5);         /* NW(8) N(1) NE(2) */
         if(direction>3)                  /* W (7)       E(3) */
            direction=direction+3;
         MoveTo(direction,x,y,spc);
      } 
      if((x!=1)&&(x!=xb)&&(y==1)) /*LeftBoundary*/
      {                                   /* N(1) NE(2) */ 
         direction=MyRandom(N,S);         /*       E(3) */
         MoveTo(direction,x,y,spc);       /* S(5) SE(4) */
      }
      if((x!=1)&&(x!=xb)&&(y==yb)) /*RightBoundary*/
      {
         direction=MyRandom(1,5);         /* NW(8) N(1) */
         if(direction!=1)                 /* W (7)      */
            direction=direction+3;        /* SW(6) S(5) */
         MoveTo(direction,x,y,spc);
      }      
   }
}


/*******************************************************************/
/* Clear the ring guard near the boundary*/
/* The data here is useless */
/*******************************************************************/
void ClearRingGuard( int spc[X_deep+2][Y_wide+2] )
{
     int i,j;
     for(j=1;j<=Y_wide;j++)
     {
         spc[0][j]=nothing;
         spc[X_deep+1][j]=nothing;
     }
     for(i=1;i<=X_deep;i++)
     {
         spc[i][0]=nothing;
         spc[i][Y_wide+1]=nothing;
     }
}

/*******************************************************************/
/* Ring guard Collision/FishLost Checking */
/*    Prev       Now                         */
/*    nothing(0) nothing(0)   OK             */
/*    nothing(0) fish   (1)   OK             */
/*    fish   (1) fish   (1)   collision      */
/*    fish   (1) nothing(0)   fish lost      */
/*******************************************************************/
void CollisionChk( int p_spc[X_deep+2][Y_wide+2],
                   int spc  [X_deep+2][Y_wide+2],
                   int mk   [X_deep+2][Y_wide+2] )
{
       int i,j,m;
       for(i=1;i<=X_deep;i++)
       {
         /*collision*/
         if((p_spc[i][1]==1)&&(spc[i][1]==1)&&(mk[i][1]==0))
            for(m=1;m<=X_deep;m++)
            {
                if(spc[m][2]==0) /*0=nothing*/
                   spc[m][2]=fish;
                break;
            }
         if((p_spc[i][Y_wide]==1)&&(spc[i][Y_wide]==1)&&(mk[i][Y_wide]==0)) 
            for(m=1;m<=X_deep;m++)
            {
                if(spc[m][Y_wide-1]==0) /*0=nothing*/
                   spc[m][Y_wide-1]=fish;
                break;
            }
         /*fish lost*/
         if((p_spc[i][1]==1)&&(spc[i][1]==0))
             spc[i][1]=fish;
         if((p_spc[i][Y_wide]==1)&&(spc[i][Y_wide]==0))
             spc[i][Y_wide]=fish;
       }
       for(j=1;j<=Y_wide;j++)
       {
         /*collision*/
         if((p_spc[1][j]==1)&&(spc[1][j]==1)&&(mk[1][j]==0))
            for(m=1;m<=Y_wide;m++)
            {
                if(spc[2][m]==0) /*0=nothing*/
                   spc[2][m]=fish;
                break;
            }
         if((p_spc[X_deep][j]==1)&&(spc[X_deep][j]==1)&&(mk[X_deep][j]==0)) 
            for(m=1;m<=Y_wide;m++)
            {
                if(spc[X_deep-1][m]==0) /*0=nothing*/
                   spc[X_deep-1][m]=fish;
                break;
            }
         /*fish lost*/
         if((p_spc[1][j]==1)&&(spc[1][j]==0))
             spc[1][j]=fish;
         if((p_spc[X_deep][j]==1)&&(spc[X_deep][j]==0))
             spc[X_deep][j]=fish;
       }

}

/*******************************************************************/
/*  Spawn mechanism                                        */
/*******************************************************************/
void Spawn(int spc[X_deep+2][Y_wide+2])
{
  int i,j,k;
  for(i=2;i<=X_deep-1;i++)
    for(j=2;j<=Y_wide-1;j++)
      if( (spc[i-1][j]==1)&&
          (spc[i+1][j]==1)&&
          (spc[i][j-1]==1)&&
          (spc[i][j+1]==1)&&
          (spc[i-1][j-1]==1)&&
          (spc[i+1][j+1]==1)&&
          (spc[i+1][j-1]==1)&&
          (spc[i-1][j+1]==1)   )
      
         spc[i][j]=fish;  
        
}

/*******************************************************************/
/*  Die mechanism                                        */
/*******************************************************************/
void Die(int spc[X_deep+2][Y_wide+2])
{
  int i,j,k;
  for(i=2;i<=X_deep-1;i++)
    for(j=2;j<=Y_wide-1;j++)
      
      if( (spc[i-1][j]==0)&&
          (spc[i+1][j]==0)&&
          (spc[i][j-1]==0)&&
          (spc[i][j+1]==0)&&
          (spc[i-1][j-1]==0)&&
          (spc[i+1][j+1]==0)&&
          (spc[i+1][j-1]==0)&&
          (spc[i-1][j+1]==0)   )
      
         spc[i][j]=fish;  
   
}


